##############################################################################
# Module : buffer.py
# Author : Old King Cole
# Date   : 06/21/2011
#
# Description : Buffer pools
#
##############################################################################

from threading import Semaphore
from math import log

READSIZE = 32768


class Buffer:
    def __init__(self, pool, size):
        self.bufferpool = pool
        self.size = size
        self.buf = memoryview(bytearray(size))

    def init(self, s = ''):
        self.length = 0
        if s:
            self.append(s)

    def append(self, s):
        l = self.length + len(s)
        self.buf[self.length:l] = s
        self.length = l

    def __len__(self):
        return self.length

    def setLength(self, length):
        self.length = length

    def __getitem__(self, index):
        return self.buf[index]

    def viewSlice(self, a = 0, b = None):
        if b is None or b > self.length:
            b = self.length
        elif b < 0:
            b += self.length
        return self.buf[a:b]

    def __getslice__(self, a, b):
        # Returns a string allowing easy external testing or converting of a small
        # part of the buffer
        # As this allocates new memory for this string it's not adequate for returning
        # a big amount of data from the buffer : in this case use viewSlice
        return self.viewSlice(a, b).tobytes()

    def release(self):
        self.bufferpool.release(self)


class CatBufferPool:
    def __init__(self, catmaxmem = 67108864):
        self.catmaxmem = catmaxmem
        self.pool = {}
        self.catsize = {}
        self.sem = Semaphore(1)
        self.overloaded = False

    def setCatMaxMem(self, catmaxmem):
        self.catmaxmem = catmaxmem

    def bufferSize(self, size, base = 32768):
        if size <= base:
            return base
        return 2 ** (int(log(size - 1, 2)) + 1)

    def catSize(self, bufsize):
        cs = self.catsize.get(bufsize)
        if cs is None:
            return 0
        return cs

    def isOverloaded(self, bufsize):
        self.sem.acquire()
        cat = self.pool.get(bufsize)
        if cat is None:
            cat = self.pool[bufsize] = []
            self.catsize[bufsize] = 0
        overloaded = (self.catsize[bufsize] - (len(cat) - 1) * bufsize > self.catmaxmem)
        if overloaded:
            self.overloaded = True
        self.sem.release()
        return overloaded

    def new(self, data = '', size = 0, checkoverload = False, base = 32768):
        if data:
            size = max(size, len(data))
        bufsize = self.bufferSize(size, base)
        self.sem.acquire()
        cat = self.pool.get(bufsize)
        if cat is None:
            cat = self.pool[bufsize] = []
            self.catsize[bufsize] = 0
        if checkoverload and self.catsize[bufsize] - (len(cat) - 1) * bufsize > self.catmaxmem:
            self.overloaded = True
            self.sem.release()
            return None
        if cat:
            x = cat.pop()
        else:
            x = Buffer(self, bufsize)
            self.catsize[bufsize] += bufsize
        self.sem.release()
        x.init(data)
        return x

    def release(self, x):
        self.pool[x.size].append(x)

    def totalSize(self):
        totalsize = 0
        self.sem.acquire()
        for bufsize in self.catsize:
            totalsize += self.catsize[bufsize]
        self.sem.release()
        return totalsize


GENBUFFERPOOL = CatBufferPool()

def newSendPiecePartBuffer(data = '', size = 0):
    return GENBUFFERPOOL.new(data = data, size = size, base = 512)

def newPieceWriteBuffer(data = '', size = 0):
    return GENBUFFERPOOL.new(data = data, size = size, base = 16384)

def newMsgPartBuffer(data = '', size = 0):
    return GENBUFFERPOOL.new(data = data, size = size, base = 512)

def newMsgBuffer(data = '', size = 0):
    return GENBUFFERPOOL.new(data = data, size = size, base = 16)

def newSendMsgBuffer(data = '', size = 0):
    return GENBUFFERPOOL.new(data = data, size = size, base = 16)

def newMetadataBuffer(data = '', size = 0):
    return GENBUFFERPOOL.new(data = data, size = size, base = 512)

def newMetadataPartBuffer(data = '', size = 0):
    return GENBUFFERPOOL.new(data = data, size = size, base = 16384)

def newReceiveBuffer(data = '', size = 0):
    return GENBUFFERPOOL.new(data = data, size = size, base = READSIZE)


PIECEREADBUFFERPOOL = CatBufferPool()

def newPieceReadBuffer(data = '', size = 0, checkoverload = False):
    return PIECEREADBUFFERPOOL.new(data = data, size = size, checkoverload = checkoverload, base = 1048576)
